import math
import pyglet
import random

import layer
import layer.audio
import layer.command
import layer.const
import layer.load
import layer.coll
import layer.task
import layer.keybind
import layer.gl

from layer.sprite import Sprite, Group, OrderedGroup, batch, Batch
from layer.entity import Entity, EntityGroup

from lagomorph.dialogue import Dialogue, DialogueDef

def dist(a, b):
    return math.sqrt((a.x - b.x) * (a.x - b.x) + (a.y - b.y) * (a.y - b.y))

class Entity(Entity):
    dfn = None
    eaten = None
    eating = None
    force_go_to = None

    @property
    def block(self):
        return (int(self.x / layer.const.BLOCK_SIZE),
                int(self.y / layer.const.BLOCK_SIZE))

    def next_day(self, world):
        pass

    def next_night(self, world):
        pass

class Interactable(layer.parse.ParseObject):
    command = ""
    radius = 10
    x = 0
    y = 0
    name = ""
    
class Map(layer.parse.ParseObject):
    load_spec = "*.map"
    interactables = []
    prompt = ""
    _shelves_ = [
        "map", "collision", "interactables", "width", "height", "prompt"]

    def find(self, name):
        for i in self.interactables:
            if i.name.lower() == name.lower():
                return i.x, i.y
        return 0, 0
    
    def __setstate__(self, d):
        super(Map, self).__setstate__(d)
        self.map_texture = layer.load.image(self.map)
        coll = layer.load.image(self.collision)
        texels = coll.image_data.get_data("RGBA", coll.width * 4)
        self.speedmap = {}
        for i in range(0, len(texels), 4):
            color = texels[i:i+4]
            v = min(color) / 255.0
            if v < 1:
                x = (i / 4) % self.width
                y = (i / 4) / self.width
                self.speedmap[(x, y)] = v

    def interact(self, x, y):
        for i in self.interactables:
            dx = i.x - x
            dy = i.y - y
            if dx * dx + dy * dy < i.radius * i.radius and i.command:
                layer.command.run(i.command)

    def prompt(self, x, y):
        for i in self.interactables:
            dx = i.x - x
            dy = i.y - y
            if dx * dx + dy * dy < i.radius * i.radius and i.command:
                return i.prompt
        return ""

    def open_grid(self, blocksize):
        grid = []
        for x in range(0, self.width, blocksize):
            l = []
            grid.append(l)
            for y in range(0, self.height, blocksize):
                for dx in range(x, x + blocksize):
                    for dy in range(y, y + blocksize):
                        if self.speed(dx, dy) < 0.1:
                            l.append(False)
                            break
                    if len(l) > y / blocksize: break
                else:
                    l.append(True)
        return grid

    def speed(self, x, y):
        return self.speedmap.get((int(x), int(y)), 1)

class World(object):
    _state = "farm"
    _dialogue = None

    def __init__(self, window):
        self.window = window
        self.rabbits = []
        self.title = Sprite(filename="title.png")
        self.diffuse_batch = Batch(z=100)
        self.farm_batch = Batch(z=-1)
        self.town_batch = Batch(z=-1)
        self.ui_batch = Batch(z=1)
        self.width = window.width
        self.height = window.height
        self.sprites = Group()
        self.ents = EntityGroup()
        self.time = TimeOfDay(self, window)
        self.player = Player(self)
        self.ents.add(self.player)
        self.door = Item(EntityDef.Find("door"), self)
        self.map = Map.Find("Farm")
        self.farm_bg = Sprite(
            z=-1, image=self.map.map_texture, batch=self.farm_batch)
        self.town_bg = Sprite(z=-1, filename="town.png", batch=self.town_batch)
        self.blocksize = layer.const.BLOCK_SIZE
        self.pathing = [[0 for y in range(self.height / self.blocksize)]
                        for x in range(self.width / self.blocksize)]
        self.open = self.map.open_grid(self.blocksize)
        self.diffsprites = []
        self._diffuse_accum = 0
        self.task = layer.task.Task(self.update)
        self.spawner = Spawner(self, "rabbit_home")
        self.night = False
        self.door.x, self.door.y = self.map.find("player_home")
        self.door.sprite.batch = None
        self.door.sprite.image.anchor_y = 0
        for x, l in enumerate(self.pathing):
            sprs = []
            for y, v in enumerate(l):
                spr = Sprite(batch=self.diffuse_batch, filename="white")
                spr.width = self.blocksize
                spr.height = self.blocksize
                spr.x = x * self.blocksize
                spr.y = y * self.blocksize
                spr.color = (255, 0, 0)
                spr.opacity = 0
                sprs.append(spr)
            self.diffsprites.append(sprs)

    def gameover(self):
        self.gameover = GameOver(self.player, self.window)
        self.disable()

    def _get_state(self):
        return self._state

    def _set_state(self, state):
        layer.keybind.pop("farm")
        layer.keybind.pop("town")
        try: self.window.batches.remove(self.farm_batch)
        except ValueError: pass
        try: self.window.batches.remove(self.town_batch)
        except ValueError: pass
        self._state = state.lower()
        if self._state == "farm":
            layer.audio.music.play("field.ogg")
            layer.keybind.push("farm")
            self.window.batches.append(self.farm_batch)
            self.task.paused = False
            self.ents.enable()
        elif self._state == "town":
            layer.audio.music.play("town.ogg")
            layer.keybind.push("town")
            self.window.batches.append(self.town_batch)
            self.task.paused = True
            self.ents.disable()

    state = property(_get_state, _set_state)

    def update(self, dt):
        if self.time.time > layer.const.EVENING:
            if self.player in self.ents:
                dt = 0 # don't advance time / diffuse
                self.player.force_go_to = self.map.find('player_home')
            elif not self.night:
                for ent in list(self.ents):
                    ent.next_night(self)
                self.spawner.spawn()
                self.night = True
                layer.audio.music.play("night.ogg")
                self.ents.add(self.door)
                self.door.sprite.batch = self.farm_batch
            self.task.scale = layer.const.EVENING_SCALE
            if self.ents.task:
                self.ents.task.scale = layer.const.EVENING_SCALE
        elif self.time.time > layer.const.MORNING:
            if self.night:
                self.night = False
                for ent in list(self.ents):
                    if isinstance(ent, Rabbit):
                        ent.force_go_to = self.map.find("rabbit_home")
                        self.night = True
                        dt = 0
            if not self.night and self.player not in self.ents:
                layer.audio.music.play("field.ogg")
                self.door.kill()
                if self.door.hp <= 0:
                    self.gameover()
                for ent in list(self.ents):
                    ent.next_day(self)
                self.ents.add(self.player)
                self.player.sprite.batch = self.farm_batch
            self.task.scale = 1
            if self.ents.task:
                self.ents.task.scale = 1
        self._diffuse_accum += dt
        if self._diffuse_accum > layer.const.REFLOW_TIMER:
            self.flow()
        self.player.inventory.update()
        self.time.update(dt)

    def dialogue(self, dialogue):
        if self._dialogue is not None:
            self._dialogue.destroy()
        self._dialogue = Dialogue(dialogue, self.window, self)
        layer.command.command(self.dialogue_advance)
        layer.command.command(self.dialogue_destroy)
        self.task.paused = True
        self.ents.disable()

    def dialogue_advance(self, choice=10000000):
        if self._dialogue:
            d = self._dialogue
            if not d.advance(choice) and d == self._dialogue:
                self.dialogue_destroy()

    def dialogue_destroy(self):
        if self._dialogue:
            self._dialogue.destroy()
            self._dialogue = None
            layer.command.uncommand(self.dialogue_advance)
            layer.command.uncommand(self.dialogue_destroy)
            self.ents.enable()
            self.task.paused = False

    def debug_diffusion(self, enable=None):
        if enable is None:
            return self.diffuse_batch in self.window.batches
        elif enable and self.diffuse_batch not in self.window.batches:
            self.window.batches.append(self.diffuse_batch)
        elif not enable and self.diffuse_batch in self.window.batches:
            self.window.batches.remove(self.diffuse_batch)

    def flow(self):
        old_path = self.pathing
        width = int(self.height / self.blocksize)
        height = int(self.height / self.blocksize)
        self.pathing = [[0.0 for y in range(height)] for x in xrange(width)]
        self.tasty = {}
        self.offensive = {}
        self.blocking = {}

        for ent in self.ents:
            if ent.dfn.tasty:
                self.tasty[ent.block] = ent
            if ent.dfn.hungry:
                self.blocking[ent.block] = ent
        
        def p(x, y):
            if (x, y) in self.blocking:
                return 0.0
            elif not self.open[x][y]:
                return 0.0
            elif (x, y) in self.tasty:
                return self.tasty[(x, y)].dfn.tasty
            elif (x, y) in self.offensive:
                return 20.0
            else:
                vleft = (x > 0) and old_path[x - 1][y] or 0
                vright = (x < width - 1) and old_path[x + 1][y] or 0
                vup = (y < height - 1) and old_path[x][y + 1] or 0
                vdown = (y > 0) and old_path[x][y - 1] or 0
                v = (vleft + vright + vup + vdown) / 4.0
                return v
        for x in xrange(width):
            for y in xrange(height):
                self.pathing[x][y] = p(x, y)
                if self.debug_diffusion():
                    self.diffsprites[x][y].opacity = (
                        self.pathing[x][y] / 60.0) * 255

    def valid_pos(self, ox, oy, nx, ny):
        dx = nx - ox
        dy = ny - oy
        err = 0
        if not (dx or dy):
            return ox, oy
        elif dx == 0 and oy < ny:
            for y in range(int(oy), int(ny) + 2, 1):
                if self.map.speed(ox, y) < 0.1:  break
                ox, oy = ox, min(y, ny)
        elif dx == 0 and oy > ny:
            for y in range(int(oy), int(ny) - 2, -1):
                if self.map.speed(ox, y) < 0.1:  break
                ox, oy = ox, max(y, ny)
        else:
            derr = (dy / dx)
            y = oy
            if dx > 0: dx, fx = 1, min
            else: dx, fx = -1, max
            if dy > 0: dy, fy = 1, min
            else: dy, fy = -1, max
            for x in range(int(ox), int(nx) + 2 * dx, dx):
                if self.map.speed(x, y) < 0.1: break
                err += derr
                ox, oy = fx(x, nx), fy(y, ny)
                if abs(err) >= 0.5:
                    y = y + dy
                    err -= 1
        return clamp(ox, 0, self.width), clamp(oy, 0, self.height)

    def interact(self):
        self.map.interact(self.player.x, self.player.y)
    
    def spawn(self, x=None, y=None):
        r = Rabbit(self, x, y)
        self.ents.add(r)

    def give_item(self, dfn, slot=0):
        poop = Item(dfn, self)
        self.player.inventory.add(poop)

    def pickup_drop(self, slot):
        if self.player.inventory[slot]:
            self.player.drop(slot)
            return
        nearest = None
        nearest_d = 1e9
        for ent in self.ents:
            if not ent.dfn.takeable: continue
            dx = self.player.x - ent.x
            dy = self.player.y - ent.y
            d = dx * dx + dy * dy
            if d < nearest_d:
                nearest = ent
                nearest_d = d
        if nearest_d <= layer.const.BLOCK_SIZE * layer.const.BLOCK_SIZE:
            self.player.pickup(nearest, slot)

    def seek(self, ent, confusion=0):
        x, y = ent.block
        confusion /= 2
        values = []
        for dx in [0, -1, 1]:
            for dy in [0, -1, 1]:
                try: values.append([self.pathing[x + dx][y + dy], dx, dy])
                except IndexError: pass
                else:
                    values[-1][2] *= 1 + random.uniform(-confusion, confusion)
        if values:
            best, dx, dy = max(values)
            if best > 0.01:
                return (dx, dy)
        return (0, 0)

    def enable(self):
        self.title.batch = None
        layer.keybind.pop("menu")
        layer.command.attr_command(self, "state")
        layer.command.attr_command(self.player, "up")
        layer.command.command(
            self.debug_diffusion, declspec=[layer.command.cmd_bool])
        layer.command.attr_command(self.player, "down")
        layer.command.attr_command(self.player, "left")
        layer.command.attr_command(self.player, "right")
        layer.command.command(self.interact)
        layer.command.command(self.pickup_drop, "pickup", declspec=[int])
        layer.command.command(self.player.drop, declspec=[int])
        layer.command.command(self.spawn, declspec=[int, int])
        layer.command.command(self.give_item, declspec=[EntityDef.Find, int])
        layer.command.command(self.dialogue, declspec=[DialogueDef.Find])
        layer.command.uncommand(self.enable)
        layer.command.command(self.disable)
        self.task.enable()
        self.ents.enable()
        self.ents.window = self.window
        self.state = 'farm'
        self.window.batches.append(self.ui_batch)

    def disable(self):
        self.title.batch = batch
        layer.keybind.push("menu")
        layer.command.command(self.enable)
        layer.command.uncommand(self.disable)
        layer.command.uncommand("up")
        layer.command.uncommand("down")
        layer.command.uncommand("left")
        layer.command.uncommand("right")
        layer.command.uncommand(self.interact)
        layer.command.uncommand(self.pickup_drop)
        layer.command.uncommand(self.player.drop)
        layer.command.uncommand(self.give_item)
        layer.command.uncommand(self.spawn)
        layer.command.uncommand(self.debug_diffusion)
        layer.command.uncommand(self.dialogue)
        self.task.disable()
        self.ents.disable()
        self.ents.window = None
        for b in (self.farm_batch, self.town_batch, self.ui_batch):
            try: self.window.batches.remove(b)
            except ValueError: pass

    def destroy(self):
        self.disable()
        self.ents.destroy()

class Spawner(object):
    def __init__(self, world, location):
        self.count = 0
        self.world = world
        self.location = location
        self.spawned = 0
        self.rabbits = []

    def spawn(self):
        x, y = self.world.map.find(self.location)
        exp = (1 + layer.const.RABBIT_SPAWN_INCREASE) ** self.count
        to_spawn = layer.const.RABBIT_INITIAL_SPAWN * exp
        while int(to_spawn) > self.spawned:
            self.rabbits.append(Rabbit(self.world, x, y))
            self.spawned += 1
        for rabbit in self.rabbits:
            self.world.ents.add(rabbit)
            rabbit.sprite.batch = self.world.farm_batch
        self.count += 1

class Inventory(object):
    class num_label(object):
        def __init__(self, number=0, x=--50, y=-50, batch=None,
                        group=None, size=48):
            self._number = number
            self._pos = [x, y]
            self._size = size
            self._symbols = {}
            self._display = []
            self.gap = -8
            self.group = group 
            self.batch = batch
            
            for symbol in range(0, 10): #+ ['minus', 'plus', 'dot', 'comma', 'e']:
                spr = Sprite(filename="num_" + str(symbol) + ".png",
                             batch=None, group=None)
                spr.image.anchor_x = spr.image.width / 2
                spr.image.anchor_y = spr.image.height / 2
                self._symbols[str(symbol)] = spr
            
            self._update()
            
        def _update(self):
            self._display = []
            digits = list(str(self._number))
            scale = 1 / (float(self._symbols['0'].height) / self._size)
            width_needed = (len(digits) * (self._symbols['0'].width     \
                            * scale + self.gap * scale)) - self.gap * scale
            x_offset = self._pos[0] - (width_needed - (self._symbols['0'].width / 2)) / 2

            n = 0
            for digit in digits:
                spr = Sprite(image=self._symbols[digit].image,
                             batch=self.batch, group=self.group)
                spr.scale = scale
                spr.x = x_offset + spr.width * n + self.gap * scale * n
                spr.y = self._pos[1]
                self._display += [spr]
                n += 1
            
        def _set_number(self, num):
            self._number = num
            self._update()
       
        def _set_size(self, new):
            self._size = new
            for symbol in self._symbols.itervalues():
                symbol.scale = spr.image.height / (float(spr.image.height) / self._size)
            self._update()

        def _set_x(self, x):
            self._pos[0] = x
            self._update()
            
        def _set_y(self, y):
            self._pos[1] = y
            self._update()
            
        number = property(lambda s: s._number, _set_number)
        size = property(lambda s: s._size, _set_size)
        x = property(lambda s: s._pos[0], _set_x)
        y = property(lambda s: s._pos[1], _set_y)
        
    def __init__(self, player, length, window):
        self.group = OrderedGroup(10)
        self.boxgroup = OrderedGroup(0, parent=self.group)
        self.sprgroup = OrderedGroup(2, parent=self.group)
        self.window = window
        self.items = []
        self.batch = batch = player.world.ui_batch
        self.bg = Sprite(group=self.boxgroup, batch=self.batch,
                         filename="top_bar.png")
        self.player = player
        self._money = 300
        self._moneylabel = Inventory.num_label(self._money, 730, 568, self.batch, self.sprgroup, 20)
        self.timeofday = {}
            
        for tod in ['twilight', 'morning', 'noon', 'afternoon', 'evening', 'midnight']:
            spr = Sprite(filename=tod + ".png", group=self.sprgroup, batch=self.batch)
            spr.opacity = 0
            spr.x = layer.const.TIME_OF_DAY_X
            spr.y = layer.const.TIME_OF_DAY_Y
            spr.image.anchor_x = spr.image.width / 2
            spr.image.anchor_y = spr.image.height / 2
            spr.scale = layer.const.TIME_OF_DAY_SCALE
            self.timeofday[tod] = spr
            
        self.time = player.world.time
        self.bg.y = self.window.height - self.bg.height
        selfplayer = player
        for i in xrange(length):
            self.items.append(None)

    def take(self, item_name):
        for i, item in enumerate(self.items):
            if item.dfn.name.lower() == item_name.lower():
                self.items[i] = None
                item.kill()
                return True
        return False

    def _get_money(self):
        return self._money
    def _set_money(self, n):
        self._moneylabel.number = self._money = n
    money = property(_get_money, _set_money)
    
    def __getitem__(self, i):
        return self.items[i]
            
    def __setitem__(self, i, item):
        self.items[i] = item

    def __delitem__(self, i):
        if self.items[i]:
            self.items[i].kill()
            self.items[i].sprite.delete()
            self[i] = None

    def add(self, item):
        for i in range(len(self.items)):
            if self.items[i] is None:
                self.player.pickup(item, i)
                break
        else:
            self.player.pickup(item, 0)

    def update(self, dt=None):
        x = layer.const.INVENTORY_X
        y = layer.const.INVENTORY_Y
        width = layer.const.INVENTORY_WIDTH
        height = layer.const.INVENTORY_HEIGHT
        
        for i, item in enumerate(self.items):
            if not item: continue
            if item.sprite.batch != self.bg.batch:
                item.sprite.batch = self.bg.batch
            if item.sprite.group != self.sprgroup:
                item.sprite.group = self.sprgroup
            item.sprite.scale = min(1, float(height) / item.sprite.image.height)
            item.sprite.x = x + i * width + width / 2
            item.sprite.y = y + height / 2
            item.sprite.opacity = 255
            if max(item.dfn.color) == 0:
                r = 128 * item.rage_delta
                g = 128 * item.hp_delta
                b = 128 * item.satiation_delta
                color = (r, g, b)
            else:
                color = item.dfn.color
            if item.sprite.color != color:
                item.sprite.color = color

        if self.time.time > 23 or self.time.time < 2:
            for sprite in self.timeofday.itervalues(): sprite.opacity = 0
            self.timeofday["midnight"].opacity = 255
        elif self.time.time < 5:
            for sprite in self.timeofday.itervalues(): sprite.opacity = 0
            self.timeofday["twilight"].opacity = 255
        elif self.time.time < 10:
            for sprite in self.timeofday.itervalues(): sprite.opacity = 0
            self.timeofday["morning"].opacity = 255
        elif self.time.time < 14:
            for sprite in self.timeofday.itervalues(): sprite.opacity = 0
            self.timeofday["noon"].opacity = 255
        elif self.time.time < 18:
            for sprite in self.timeofday.itervalues(): sprite.opacity = 0
            self.timeofday["afternoon"].opacity = 255
        elif self.time.time < 23:
            for sprite in self.timeofday.itervalues(): sprite.opacity = 0
            self.timeofday["evening"].opacity = 255

class Player(Entity):
    saved = 0
    killed = 0
    grown = 0
    talked_to = {   'tewi': False,
                    'lapine': False,
                    'conilia': False
                }
    affection = {   'tewi': 0.3,
                    'lapine': 0.3,
                    'conilia': 0.3
                }
    def __init__(self, world):
        super(Player, self).__init__()
        self.dfn = EntityDef.Find("player")
        self.up = 0
        self.down = 0
        self.left = 0
        self.right = 0
        self.x = world.width / 2
        self.y = world.height / 2
        self.world = world
        self.images = self.dfn.down_anim
        self.sprite = Sprite(image=self.images[0], group=world.sprites,
                             batch=world.farm_batch)
        self.inventory = Inventory(self, 8, world.window)
        self.sprite.color = self.dfn.color
        self.distance = 0
        self.label = pyglet.text.Label(
            text="", group=self.sprite.group, batch=self.sprite.batch,
            anchor_x="center")
        self.label.font_size = 16
        self.dx = 0
        self.dy = 0
        self.days = 0

    def next_day(self, world):
        self.days += 1
        if self.days > 30:
            self.world.gameover()

    def pickup(self, item, slot):
        self.drop(slot)
        self.world.ents.kill(item)
        self.inventory[slot] = item

    def drop(self, slot):
        item = self.inventory[slot]
        if item:
            item.x = self.x
            item.y = self.y
            item.sprite.group = self.world.sprites
            item.sprite.color = item.dfn.color
            item.sprite.batch = self.world.farm_batch
            item.sprite.scale = (item.dfn.hp
                                 and (float(item.hp) / item.dfn.hp)
                                 or 1)
            self.world.ents.add(item)
        self.inventory[slot] = None

    def tick(self, dt):
        if self.force_go_to:
            left = self.force_go_to[0]  < self.x
            right = self.force_go_to[0]  > self.x
            down = self.force_go_to[1]  < self.y
            up = self.force_go_to[1]  > self.y
            dx = self.x - self.force_go_to[0]
            dy = self.y - self.force_go_to[1]
            pre_dist = dx * dx + dy * dy

        else:
            left = self.left
            right = self.right
            up = self.up
            down = self.down
        self.dx = right - left
        self.dy = up - down
        if self.dx or self.dy:
            l = math.sqrt(self.dx * self.dx + self.dy * self.dy)
            speed = (self.dfn.speed * self.world.map.speed(self.x, self.y))
            if speed == 0 and self.force_go_to:
                speed = self.dfn.speed / 3
            dx = dt * speed * self.dx / l
            dy = dt * speed * self.dy / l
            nx = self.x + dx
            ny = self.y + dy
            if not self.force_go_to:
                nx, ny = self.world.valid_pos(self.x, self.y, nx, ny)
            dist = math.sqrt((nx - self.x) * (nx - self.x) +
                             (ny - self.y) * (ny - self.y))
            self.x = nx
            self.y = ny
            self.distance += dist

        if self.force_go_to:
            dx = self.x - self.force_go_to[0]
            dy = self.y - self.force_go_to[1]
            if pre_dist < dx * dx + dy * dy or pre_dist < 1:
                self.x, self.y = self.force_go_to
                self.force_go_to = None
                self.kill()
            
    def draw(self):
        if self.dy < 0: self.images = self.dfn.down_anim
        elif self.dy > 0: self.images = self.dfn.up_anim
        elif self.dx < 0: self.images = self.dfn.left_anim
        elif self.dx > 0: self.images = self.dfn.right_anim
        self.label.text = self.world.map.prompt(self.x, self.y)
        self.label.x = self.x
        self.label.y = self.y - 20

        imgidx = int(
            (self.distance / self.dfn.anim_distance) % len(self.images))
        img = self.images[imgidx]
        if img is not self.sprite.image:
            self.sprite.image = img
        if self.sprite.color != self.dfn.color:
            self.sprite.color = self.dfn.color
        if self.sprite.position != (self.x, self.y):
            self.sprite.position = (self.x, self.y)

class TimeOfDay(object):
    def __init__(self, world, window):
        self.group = OrderedGroup(100)
        self.dark = Sprite(
            filename="white", group=OrderedGroup(0, parent=self.group),
            batch=world.farm_batch)
        self.light = Sprite(
            filename="white", group=OrderedGroup(1, parent=self.group),
            batch=world.farm_batch,
            blend_src=layer.gl.GL_SRC_ALPHA,
            blend_dest=layer.gl.GL_ONE)
        self.dark.x = self.light.x = 0
        self.dark.y = self.light.y = 0
        self.dark.width = self.light.width = window.width
        self.dark.height = self.light.height = window.height
        self.dark.color = (0, 0, 0)
        self.light.color = (255, 255, 0)
        self.time = 8
        self.light.opacity = 0

    def update(self, dt):
        hours = 24 * dt / layer.const.DAY_LENGTH
        self.time = (self.time + hours) % 24
        if self.time > 22 or self.time < 2:
            self.dark.opacity = 128
        elif self.time > 18:
            p = (self.time - 18) / 4.0
            self.dark.opacity4 = [128 * p,
                                  min(128, 256 * p),
                                  min(128, 256 * p),
                                  128 * p,
                                  ]
        elif self.time < 6:
            p = 1 - (self.time - 2) / 4.0
            self.dark.opacity4 = [min(128, 256 * p),
                                  128 * p,
                                  128 * p,
                                  min(128, 256 * p),
                                  ]
        else:
            self.dark.opacity = 0
        if 4 < self.time < 8:
            # light fades black to yellow, additive, but also fades out
            self.light.opacity = 255 * (self.time - 4) / 4
            p = (self.time - 4) / 4.0
            ceast = (255 * p, 255 * p, 0)
            cwest = (512 * p, 512 * p, 0)
            self.light.color4 = [ceast, cwest, cwest, ceast]
            self.light.opacity = 64 * (1 - p)
        else:
            self.light.opacity = 0
            
class Rabbit(Entity):
    rage = 1

    def __init__(self, world, x=None, y=None):
        self.world = world
        self.x = x or world.width / 2
        self.y = y or world.height / 2
        self.dfn = EntityDef.Find("Rabbit")
        self.images = self.dfn.down_anim
        self.sprite = Sprite(image=self.images[0], group=world.sprites,
                             batch=world.farm_batch)
        self.sprite.color = (255, 255, 255)
        self.distance = 0
        self.dx = 0
        self.dy = 0
        self.full = 0
        self.poop_timer = 0
        self.hp = self.dfn.hp
        self.poop = []

    def tick(self, dt):
        if self.poop_timer >= 0:
            self.poop_timer = random.gauss(
                layer.const.POOP_MEAN, layer.const.POOP_VARIANCE)
            if len(self.poop):
                poop = self.poop.pop()
                self.world.ents.add(poop)
                poop.sprite.batch = self.world.farm_batch
                poop.x = self.x
                poop.y = self.y
            
        if self.force_go_to:
            if self.eating:
                self.eating.eaten = None
                self.eating = None

            left = self.force_go_to[0]  < self.x
            right = self.force_go_to[0]  > self.x
            down = self.force_go_to[1]  < self.y
            up = self.force_go_to[1]  > self.y
            dx = self.x - self.force_go_to[0]
            dy = self.y - self.force_go_to[1]
            pre_dist = dx * dx + dy * dy
            dx = right - left
            dy = up - down
        elif not self.eating:
            dx, dy = self.world.seek(
                self, confusion=layer.const.RABBIT_CONFUSION)
        elif self.eating:
            hp = self.dfn.damage * dt * self.rage
            self.eating.hp -= hp
            self.full += hp * (layer.const.HUNGER_HP_RATIO *
                               self.eating.satiation_delta)
            if self.eating.hp < 0:
                self.hp += self.eating.dfn.hp * -(self.eating.hp_delta - 1)
                self.eating.kill()
                self.rage *= min(1, self.eating.rage_delta)
                poop = Item(EntityDef.Find("poop"), self.world)
                poop.rage_delta = self.eating.rage_delta
                poop.hp_delta = self.eating.hp_delta
                poop.satiation_delta = self.eating.satiation_delta
                poop.cross(poop)
                self.eating = None
                self.poop.append(poop)
            if self.hp < 0:
                dead = Item(EntityDef.Find("dead_rabbit"), self.world)
                self.world.ents.add(dead)
                dead.sprite.batch = self.world.farm_batch
                dead.x = self.x
                dead.y = self.y
                self.kill()
                self.world.player.saved += 1
            if self.rage < 0.6:
                dead = Item(EntityDef.Find("cured_rabbit"), self.world)
                self.world.ents.add(dead)
                dead.sprite.batch = self.world.farm_batch
                dead.x = self.x
                dead.y = self.y
                self.world.player.killed += 1
                self.kill()
            dx = 0
            dy = 0
        self.dx = dx
        self.dy = dy
        if dx or dy:
            speed = (self.dfn.speed * self.world.map.speed(self.x, self.y))
            speed *= self.rage
            if speed == 0 and self.force_go_to:
                speed = self.dfn.speed / 3
            dx *= speed * dt
            dy *= speed * dt
            self.distance += math.sqrt(dx * dx + dy * dy)
            self.x += dx
            self.y += dy

        if not self.force_go_to:
            for ent in self.world.ents:
                if ent is self: continue
                if ent.eaten: continue
                if self.full > 0: continue
                if ent.dfn and ent.dfn.tasty and self.block == ent.block:
                    ent.eaten = self
                    self.eating = ent
                    self.full = 0
            self.full -= dt
        else:
            dx = self.x - self.force_go_to[0]
            dy = self.y - self.force_go_to[1]
            if pre_dist < dx * dx + dy * dy or pre_dist < 1:
                self.x, self.y = self.force_go_to
                self.force_go_to = None
                self.kill()

    def draw(self):
        if not ((self.images == self.dfn.down_anim and self.dy < 0)
                or (self.images == self.dfn.up_anim and self.dy > 0)
                or (self.images == self.dfn.left_anim and self.dx < 0)
                or (self.images == self.dfn.right_anim and self.dx > 0)):
            if self.dy < 0: self.images = self.dfn.down_anim
            elif self.dy > 0: self.images = self.dfn.up_anim
            elif self.dx < 0: self.images = self.dfn.left_anim
            elif self.dx > 0: self.images = self.dfn.right_anim

        if self.dx or self.dy:
            imgidx = int(
                (self.distance / self.dfn.anim_distance) % len(self.images))
        else:
            imgidx = 3
        img = self.images[imgidx]

        if img is not self.sprite.image:
            self.sprite.image = img
        if self.sprite.color != self.dfn.color:
            self.sprite.color = self.dfn.color
        if self.sprite.position != (self.x, self.y):
            self.sprite.position = (self.x, self.y)

        self.sprite.x = self.x
        self.sprite.y = self.y

class EntityDef(layer.parse.ParseObject):
    load_spec = "*.item"
    _thawers_ = dict(color=layer.colors.value)
    _freezers_ = dict(color=layer.colors.name)
    tasty = False
    image = None
    down_anim = []
    up_anim = []
    left_anim = []
    right_anim = []
    anchor_x = None
    anchor_y = None
    hungry = False
    takeable = False
    speed = 100
    anim_distance = 10
    hp = 0
    damage = 0
    color = (255, 255, 255)
    growth = []
    days_to_grow = 1e9
    grows_into = []
    poops = []
    can_fertilize = False
    can_be_fertilized = False

    def __getstate__(self):
        d = super(EntityDef, self).__getstate__()
        try: d["image"] = d["image"].filename
        except KeyError: pass
        except AttributeError: del(d["image"])

        for attr in ["down_anim", "up_anim", "left_anim", "right_anim",
                     "growth"]:
            try: d[attr] = map(lambda i: i.filename, d[attr])
            except KeyError: pass
            except AttributeError: del(d[attr])
        return d
    
    def __setstate__(self, d):
        super(EntityDef, self).__setstate__(d)
        if self.image:
            self.image = layer.load.image(self.image)
            self.image.anchor_x = self.anchor_x or self.image.width / 2
            self.image.anchor_y = self.anchor_y or self.image.height / 2
        for attr in ["down_anim", "up_anim", "left_anim", "right_anim",
                     "growth"]:
            l = getattr(self, attr)
            if l:
                l = map(layer.load.image, l)
                setattr(self, attr, l)
                for image in l:
                    image.anchor_x = self.anchor_x or image.width / 2
                    image.anchor_y = self.anchor_y or image.height / 2

class Item(Entity):
    hp_delta = 1
    rage_delta = 1
    satiation_delta = 1
    
    def __init__(self, dfn, world):
        super(Item, self).__init__()
        self.dfn = dfn
        self.hp = self.dfn.hp
        self.sprite = Sprite(
            image=self.dfn.image, group=None, batch=world.farm_batch)
        self.sprite.color = self.dfn.color
        self.days = 0

    def cross(self, other):
        for attr in ('hp_delta', 'rage_delta', 'satiation_delta'):
            v1 = getattr(self, attr)
            v2 = getattr(other, attr)
            v1, v2 = min(v1, v2), max(v1, v2)
            v1 -= 0.3
            v2 += 0.3
            v = random.uniform(v1, v2)
            setattr(self, attr, clamp(v, 0, 2))

    def next_night(self, world):
        if self.dfn.can_be_fertilized:
            for ent in world.ents:
                if (ent.dfn.can_fertilize
                    and dist(ent, self) < world.blocksize * 2):
                    self.cross(ent)
                    ent.kill()
            
    def next_day(self, world):
        self.days += 1
        if self.days >= self.dfn.days_to_grow:
            grows_into = random.choice(self.dfn.grows_into)
            plant = Item(EntityDef.Find(grows_into), world)
            plant.x = self.x
            plant.y = self.y
            plant.hp_delta = self.hp_delta
            plant.rage_delta = self.rage_delta
            plant.satiation_delta = self.satiation_delta
            world.ents.add(plant)
            world.player.grown += 1
            self.kill()
        elif self.days < len(self.dfn.growth):
            self.sprite.batch = None
            self.sprite.image = self.dfn.growth[self.days]
            self.sprite.batch = world.farm_batch
            
    def draw(self):
        try: img = self.dfn.growth[self.days]
        except IndexError: img = self.dfn.image
        if max(self.dfn.color) == 0:
            r = 128 * self.rage_delta
            g = 128 * self.hp_delta
            b = 128 * self.satiation_delta
            color = (r, g, b)
        else:
            color = self.dfn.color
            
        if img != self.sprite.image:
            self.sprite.image = img
        if self.sprite.color != color:
            self.sprite.color = color
        if self.sprite.position != (self.x, self.y):
            self.sprite.position = (self.x, self.y)
        if self.dfn.hp and self.sprite.scale != (float(self.hp) / self.dfn.hp):
            self.sprite.scale = (float(self.hp) / self.dfn.hp)

class GameOver(object):
    def __init__(self, player, window):
        text = """<center><font color=white>
<font color=white>You:<br>\
<font color=white>Survived %(days)d days<br>\
<font color=white>Earned $%(money)d<br>\
<font color=white>Killed %(killed)d rabbits<br>\
<font color=white>Cured %(cured)d rabbits<br>\
<font color=white>Grew %(grown)d plants</font></center>"""
        text = text % dict(
            money=player.inventory.money,
            days=player.days,
            killed=player.killed,
            cured=player.saved,
            grown=player.grown)
        self.batch = Batch(z=2)
        self.window = window
        self.bg = Sprite(filename="gameover.png", batch=self.batch,
                     group=self.batch.Group(1))
        self.label = pyglet.text.HTMLLabel(            
            text, batch=self.batch,
            y=window.height * 0.6,
            multiline=True,
            width=window.width / 2,
            group=self.batch.Group(2))
        self.label.font_size = 20
        window.batches.append(self.batch)

    def delete(self):
        self.window.batches.remove(self.batch)
        self.bg.delete()
        self.label.delete()
            
def init(argv):
    window = layer.init(
        "lagomorph", _("I Am Lagomorph"), argv, size=(800, 600))
    return window

def main(window, argv):
    layer.command.process_argv(argv)
    layer.keybind.push("menu")
    layer.keybind.push("first")
    world = World(window)
    layer.command.command(world.enable)
    layer.audio.music.play("town.ogg")
    layer.run(window)
